from unittest.mock import Mock, patch

import os

import numpy as np
import qp

import rail.evaluation.metrics.pointestimates as pe
from rail.core.data import QPHandle, TableHandle,  QPOrTableHandle
from rail.core.stage import RailStage
from rail.utils.path_utils import find_rail_file
from rail.evaluation.evaluator import OldEvaluator
from rail.evaluation.dist_to_dist_evaluator import DistToDistEvaluator
from rail.evaluation.dist_to_point_evaluator import DistToPointEvaluator
from rail.evaluation.point_to_point_evaluator import PointToPointEvaluator
from rail.evaluation.single_evaluator import SingleEvaluator

# values for metrics
OUTRATE = 0.0
KSVAL = 0.367384
CVMVAL = 20.63155
ADVAL_ALL = 82.51480
ADVAL_CUT = 1.10750
CDEVAL = -4.31200
SIGIQR = 0.0045947
BIAS = -0.00001576
OUTRATE = 0.0
SIGMAD = 0.0046489


def _get_files():
    possible_local_file = './examples_data/evaluation_data/data/output_fzboost.hdf5'
    if os.path.exists(possible_local_file):
        pdfs_file = os.path.abspath(possible_local_file)
    else:
        pdfs_file = 'examples_data/evaluation_data/data/output_fzboost.hdf5'
        try:
            os.makedirs(os.path.dirname(pdfs_file))
        except FileExistsError:
            pass
        curl_com = f"curl -o {pdfs_file} https://portal.nersc.gov/cfs/lsst/PZ/output_fzboost.hdf5"
        os.system(curl_com)
    ztrue_file = find_rail_file('examples_data/testdata/test_dc2_validation_9816.hdf5')
    return pdfs_file, ztrue_file

    
def construct_test_ensemble():
    np.random.seed(87)
    nmax = 2.5
    NPDF = 399
    true_zs = np.random.uniform(high=nmax, size=NPDF)
    locs = np.expand_dims(true_zs + np.random.normal(0.0, 0.01, NPDF), -1)
    true_ez = (locs.flatten() - true_zs) / (1.0 + true_zs)
    scales = np.ones((NPDF, 1)) * 0.1 + np.random.uniform(size=(NPDF, 1)) * 0.05
    n_ens = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))  # pylint: disable=no-member
    zgrid = np.linspace(0, nmax, 301)
    grid_ens = n_ens.convert_to(qp.interp_gen, xvals=zgrid)
    return zgrid, true_zs, grid_ens, true_ez


def test_point_metrics():
    zgrid, zspec, pdf_ens, true_ez = construct_test_ensemble()
    zb = pdf_ens.mode(grid=zgrid).flatten()

    ez = pe.PointStatsEz(zb, zspec).evaluate()
    assert np.allclose(ez, true_ez, atol=1.0e-2)
    # grid limits ez vals to ~10^-2 tol

    sig_iqr = pe.PointSigmaIQR(zb, zspec).evaluate()
    assert np.isclose(sig_iqr, SIGIQR)

    bias = pe.PointBias(zb, zspec).evaluate()
    assert np.isclose(bias, BIAS)

    out_rate = pe.PointOutlierRate(zb, zspec).evaluate()
    assert np.isclose(out_rate, OUTRATE)

    sig_mad = pe.PointSigmaMAD(zb, zspec).evaluate()
    assert np.isclose(sig_mad, SIGMAD)


def test_evaluation_stage():
    DS = RailStage.data_store
    _zgrid, zspec, pdf_ens, _true_ez = construct_test_ensemble()
    pdf = DS.add_data("pdf", pdf_ens, QPHandle)
    truth_table = dict(redshift=zspec)
    truth = DS.add_data("truth", truth_table, TableHandle)
    evaluator = OldEvaluator.make_stage(name="Eval")
    evaluator.evaluate(pdf, truth)

    os.remove(
        evaluator.get_output(evaluator.get_aliased_tag("output"), final_name=True)
    )


def test_dist_to_dist_evaluator():
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    pdfs_file, ztrue_file = _get_files()
    stage_dict = dict(
        #metrics=['cvm', 'ks', 'rmse', 'kld', 'ad'],
        metrics=['rmse'],
        _random_state=None,
    )

    ensemble = DS.read_file(key='pdfs_data', handle_class=QPHandle, path=pdfs_file)

    dtd_stage = DistToDistEvaluator.make_stage(name='dist_to_dist', **stage_dict)
    dtd_stage_single = DistToDistEvaluator.make_stage(name='dist_to_dist_single', force_exact=True, **stage_dict)

    dtd_results = dtd_stage.evaluate(ensemble, ensemble)
    dtd_results_single = dtd_stage_single.evaluate(ensemble, ensemble)


def test_dist_to_point_evaluator():
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    pdfs_file, ztrue_file = _get_files()
    stage_dict = dict(
        metrics=['cdeloss', 'pit', 'brier'],
        _random_state=None,
        metric_config={
            'brier': {'limits':(0,3.1)},
        }        
    )

    ensemble = DS.read_file(key='pdfs_data', handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file('ztrue_data', TableHandle, ztrue_file)

    dtp_stage = DistToPointEvaluator.make_stage(name='dist_to_point', **stage_dict)
    dtp_stage_single = DistToPointEvaluator.make_stage(name='dist_to_point_single', force_exact=True, **stage_dict)

    dtp_results = dtp_stage.evaluate(ensemble, ztrue_data)
    dtp_results_single = dtp_stage_single.evaluate(ensemble, ztrue_data)

    qp_dict = qp.read_dict(dtp_results_single['single_distribution_summary'].path)
    assert qp_dict['pit'].npdf == 1

    qp_dict_handle = dtp_results_single['single_distribution_summary']
    qp_dict_handle.data = None
    qp_dict_2 = qp_dict_handle.read()
    assert qp_dict_2['pit'].npdf == 1

    qp_dict_handle.data = None

    with qp_dict_handle.open() as qp_dict_data:
        assert qp_dict_data
    
    
def test_point_to_point_evaluator():
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    pdfs_file, ztrue_file = _get_files()
    stage_dict = dict(
        metrics=['point_stats_ez', 'point_stats_iqr', 'point_bias', 'point_outlier_rate', 'point_stats_sigma_mad'],
        _random_state=None,
        hdf5_groupname='photometry',
        point_estimate_key='zmode',
        chunk_size=10000,
    )

    ensemble = DS.read_file(key='pdfs_data', handle_class=QPHandle, path=pdfs_file)
    ztrue_data = DS.read_file('ztrue_data', TableHandle, ztrue_file)

    ptp_stage = PointToPointEvaluator.make_stage(name='point_to_point', **stage_dict)
    ptp_stage_single = PointToPointEvaluator.make_stage(name='point_to_point_single', force_exact=True, **stage_dict)
    
    ptp_results = ptp_stage.evaluate(ensemble, ztrue_data)
    ptp_results_single = ptp_stage_single.evaluate(ensemble, ztrue_data)



def test_single_evaluator():
    DS = RailStage.data_store
    DS.__class__.allow_overwrite = True
    pdfs_file, ztrue_file = _get_files()
    stage_dict = dict(
        metrics=['cvm', 'ks', 'omega', 'kld', 'cdeloss', 'point_stats_ez', 'point_stats_iqr'],
        _random_state=None,
        hdf5_groupname='photometry',
        point_estimates=['zmode'],
        truth_point_estimates=['redshift'],
        chunk_size=1000,
    )
    ensemble_d = DS.add_data('pdfs_data_2', None, QPOrTableHandle, path=pdfs_file)
    ztrue_data_d = DS.add_data('ztrue_data_2', None, QPOrTableHandle, path=ztrue_file)

    single_stage = SingleEvaluator.make_stage(name='single', **stage_dict)
    single_stage_single = SingleEvaluator.make_stage(name='single_single', force_exact=True, **stage_dict)

    single_results = single_stage.evaluate(ensemble_d, ztrue_data_d)
    single_results_single = single_stage_single.evaluate(ensemble_d, ztrue_data_d)



"""
A summarizer that simple makes a histogram of a point estimate
"""

import numpy as np
import qp

from ceci.config import StageParameter as Param
from rail.estimation.summarizer import PZSummarizer
from rail.estimation.informer import PzInformer
from rail.core.data import QPHandle


class NaiveStackInformer(PzInformer):
    """Placeholder Informer"""

    name = "NaiveStackInformer"
    config_options = PzInformer.config_options.copy()

    def __init__(self, args, comm=None):
        PzInformer.__init__(self, args, comm=comm)

    def run(self):
        self.add_data("model", np.array([None]))


class NaiveStackSummarizer(PZSummarizer):
    """Summarizer which stacks individual P(z)"""

    name = "NaiveStackSummarizer"
    config_options = PZSummarizer.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        nsamples=Param(int, 1000, msg="Number of sample distributions to create"),
    )
    inputs = [("input", QPHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def __init__(self, args, comm=None):
        PZSummarizer.__init__(self, args, comm=comm)
        self.zgrid = None

    def run(self):
        iterator = self.input_iterator("input")
        self.zgrid = np.linspace(
            self.config.zmin, self.config.zmax, self.config.nzbins + 1
        )
        # Initiallizing the stacking pdf's
        yvals = np.zeros((1, len(self.zgrid)))
        bvals = np.zeros((self.config.nsamples, len(self.zgrid)))
        bootstrap_matrix = self._broadcast_bootstrap_matrix()

        first = True
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s} - {e}")
            self._process_chunk(s, e, test_data, first, bootstrap_matrix, yvals, bvals)
            first = False
        if self.comm is not None:  # pragma: no cover
            bvals, yvals = self._join_histograms(bvals, yvals)

        if self.rank == 0:
            sample_ens = qp.Ensemble(
                qp.interp, data=dict(xvals=self.zgrid, yvals=bvals)
            )
            qp_d = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=yvals))
            self.add_data("output", sample_ens)
            self.add_data("single_NZ", qp_d)

    def _process_chunk(self, start, end, data, _first, bootstrap_matrix, yvals, bvals):
        pdf_vals = data.pdf(self.zgrid)
        yvals += np.expand_dims(
            np.sum(np.where(np.isfinite(pdf_vals), pdf_vals, 0.0), axis=0), 0
        )
        # qp_d is the normalized probability of the stack, we need to know how many galaxies were
        for i in range(self.config.nsamples):
            bootstrap_draws = bootstrap_matrix[:, i]
            # Neither all of the bootstrap_draws are in this chunk nor the index starts at "start"
            mask = (bootstrap_draws >= start) & (bootstrap_draws < end)
            bootstrap_draws = bootstrap_draws[mask] - start
            bvals[i] += np.sum(pdf_vals[bootstrap_draws], axis=0)



# pylint: disable=no-member
import pytest
import numpy as np

import qp
from rail.estimation.estimator import CatEstimator


def test_custom_point_estimate():
    """This test checks to make sure that the inheritance mechanism is working
    for child classes of `CatEstimator`.
    """

    MEANING_OF_LIFE = 42.0

    class TestEstimator(CatEstimator):
        name = "TestEstimator"

        def __init__(self, args, comm=None):
            CatEstimator.__init__(self, args, comm=comm)

        def _calculate_mode_point_estimate(self, qp_dist=None, grid=None):
            return np.ones(100) * MEANING_OF_LIFE

    config_dict = {"calculated_point_estimates": ["mode"]}

    test_estimator = TestEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))  # pylint: disable=no-member

    result = test_estimator.calculate_point_estimates(test_ensemble)

    assert np.all(result.ancil["mode"] == MEANING_OF_LIFE)


def test_basic_point_estimate():
    """This test checks to make sure that all the basic point estimates are
    executed when requested in the configuration dictionary.
    """

    config_dict = {
        "calculated_point_estimates": ["mean", "median", "mode"],
        "zmin": 0.0,
        "zmax": 3.0,
        "nzbins": 301,
    }

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))  # pylint: disable=no-member
    result = test_estimator.calculate_point_estimates(test_ensemble, None)

    # note: we're not interested in testing the values of point estimates,
    # just that they were added to the ancillary data.
    assert "mode" in result.ancil
    assert "median" in result.ancil
    assert "mean" in result.ancil


def test_mode_no_grid():
    """This exercises the KeyError logic in `_calculate_mode_point_estimate`."""
    config_dict = {"zmin": 0.0, "nzbins": 100, "calculated_point_estimates": ["mode"]}

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    with pytest.raises(KeyError) as excinfo:
        _ = test_estimator.calculate_point_estimates(None, None)

    assert "to be defined in stage configuration" in str(excinfo.value)


def test_mode_no_point_estimates():
    """This exercises the KeyError logic in `_calculate_mode_point_estimate`."""
    config_dict = {"zmin": 0.0, "nzbins": 100}

    test_estimator = CatEstimator.make_stage(name="test", **config_dict)

    locs = 2 * (np.random.uniform(size=(100, 1)) - 0.5)
    scales = 1 + 0.2 * (np.random.uniform(size=(100, 1)) - 0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))  # pylint: disable=no-member

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert output_ensemble.ancil is None

def test_keep_existing_ancil_data():
    """Make sure that we don't overwrite the ancil data if it already exists.
    """
    config_dict = {'zmin':0.0, 'zmax': 3.0, 'nzbins':100, 'calculated_point_estimates': ['mode']}

    test_estimator = CatEstimator.make_stage(name='test', **config_dict)

    locs = 2* (np.random.uniform(size=(100,1))-0.5)
    scales = 1 + 0.2*(np.random.uniform(size=(100,1))-0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))

    test_ensemble.set_ancil({'foo': np.zeros(100)})

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert 'foo' in output_ensemble.ancil
    assert np.all(output_ensemble.ancil['foo'] == 0.0)
    assert len(output_ensemble.ancil['foo']) == 100

def test_write_new_ancil_data():
    """Make sure that we don't overwrite the ancil data if it already exists.
    """
    config_dict = {'zmin':0.0, 'zmax': 3.0, 'nzbins':100, 'calculated_point_estimates': ['mode']}

    test_estimator = CatEstimator.make_stage(name='test', **config_dict)

    locs = 2* (np.random.uniform(size=(100,1))-0.5)
    scales = 1 + 0.2*(np.random.uniform(size=(100,1))-0.5)
    test_ensemble = qp.Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))

    test_ensemble.set_ancil({'foo': np.zeros(100)})

    output_ensemble = test_estimator.calculate_point_estimates(test_ensemble, None)

    assert 'mode' in output_ensemble.ancil
    assert len(output_ensemble.ancil['mode']) == 100




# This is a deprecative path for importing RAILDIR, it is only for the transition period
# to move RAILDIR and find_rail_file to utils.path_utils

from rail.utils.path_utils import find_rail_file, RAILDIR

find_rail_file = find_rail_file


"""
A summarizer that simple makes a histogram of a point estimate
"""

import numpy as np
import qp

from ceci.config import StageParameter as Param
from rail.estimation.summarizer import PZSummarizer
from rail.estimation.informer import PzInformer
from rail.core.data import QPHandle


class PointEstHistInformer(PzInformer):
    """Placeholder Informer"""

    name = "PointEstHistInformer"
    config_options = PzInformer.config_options.copy()

    def __init__(self, args, comm=None):
        PzInformer.__init__(self, args, comm=comm)

    def run(self):
        self.add_data("model", np.array([None]))


class PointEstHistSummarizer(PZSummarizer):
    """Summarizer which simply histograms a point estimate"""

    name = "PointEstHistSummarizer"
    config_options = PZSummarizer.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        point_estimate=Param(str, "zmode", msg="Which point estimate to use"),
        nsamples=Param(int, 1000, msg="Number of sample distributions to return"),
    )
    inputs = [("input", QPHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def __init__(self, args, comm=None):
        PZSummarizer.__init__(self, args, comm=comm)
        self.zgrid = None
        self.bincents = None

    def run(self):
        iterator = self.input_iterator("input")
        self.zgrid = np.linspace(
            self.config.zmin, self.config.zmax, self.config.nzbins + 1
        )
        self.bincents = 0.5 * (self.zgrid[1:] + self.zgrid[:-1])
        bootstrap_matrix = self._broadcast_bootstrap_matrix()
        # Initiallizing the histograms
        single_hist = np.zeros(self.config.nzbins)
        hist_vals = np.zeros((self.config.nsamples, self.config.nzbins))

        first = True
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s} - {e}")
            self._process_chunk(
                s, e, test_data, first, bootstrap_matrix, single_hist, hist_vals
            )
            first = False
        if self.comm is not None:  # pragma: no cover
            hist_vals, single_hist = self._join_histograms(hist_vals, single_hist)

        if self.rank == 0:
            sample_ens = qp.Ensemble(
                qp.hist, data=dict(bins=self.zgrid, pdfs=np.atleast_2d(hist_vals))
            )
            qp_d = qp.Ensemble(
                qp.hist, data=dict(bins=self.zgrid, pdfs=np.atleast_2d(single_hist))
            )
            self.add_data("output", sample_ens)
            self.add_data("single_NZ", qp_d)

    def _process_chunk(
        self, start, end, test_data, _first, bootstrap_matrix, single_hist, hist_vals
    ):
        zb = test_data.ancil[self.config.point_estimate]
        single_hist += np.histogram(zb, bins=self.zgrid)[0]
        for i in range(self.config.nsamples):
            bootstrap_indeces = bootstrap_matrix[:, i]
            # Neither all of the bootstrap_draws are in this chunk nor the index starts at "start"
            mask = (bootstrap_indeces >= start) & (bootstrap_indeces < end)
            bootstrap_indeces = bootstrap_indeces[mask] - start
            zarr = zb[bootstrap_indeces]
            hist_vals[i] += np.histogram(zarr, bins=self.zgrid)[0]



"""Abstract base class defining a noisifier.

The key feature here is the run adds noise to the catalog. 
Intended subclasses are noisifier that adds LSST noise / other telescope noise
"""

from rail.core.stage import RailStage
from rail.core.data import PqHandle
from ceci.config import StageParameter as Param

class Noisifier(RailStage):
    """Base class Noisifier, which adds noise to the input catalog

    Noisifier take "input" data in the form of pandas dataframes in Parquet 
    files and provide as "output" another pandas dataframes written to Parquet 
    files.
    """

    name = 'Noisifier'
    config_options = RailStage.config_options.copy()
    config_options.update(
        seed=Param(default=None, required=False, msg="Set to an `int` to force reproducible results."),
    )
    inputs = [('input', PqHandle)]
    outputs = [('output', PqHandle)]

    def __init__(self, args, comm=None):
        """Initialize Noisifier that can add noise to photometric data"""
        RailStage.__init__(self, args, comm=comm)
        
        
    def _initNoiseModel(self):  # pragma: no cover
        raise NotImplementedError("Noisifier._initNoiseModel()")
        
    def _addNoise(self):  # pragma: no cover
        raise NotImplementedError("Noisifier._addNoise()")

    def __call__(self, sample, seed: int = None):
        """The main interface method for ``Noisifier``.

        Adds noise to the input catalog

        This will attach the input to this `Noisifier` 

        Then it will call the _initNoiseModel() and _addNoise(), which need to be
        implemented by the sub-classes.

        The _initNoiseModel() method will initialize the noise model of the sub-classes, and 
        store the noise model as self.noiseModel
        
        The _addNoise() method will add noise to the flux and magnitude of the column of the
        catalog. 
        
        The finalize() method will check the end results (like preserving number of rows)

        Finally, this will return a PqHandle providing access to that output 
        data.

        Parameters
        ----------
        sample : table-like
            The sample to be degraded
        seed : int, default=None
            An integer to set the numpy random seed

        Returns
        -------
        output_data : PqHandle
            A handle giving access to a table with degraded sample
        """
        if seed is not None:
            self.config.seed = seed
        self.set_data('input', sample)
        
        self.run()
        self.finalize()
        return self.get_handle('output')


    def run(self):
        
        self._initNoiseModel()        
        self._addNoise()
        
        
